@@ -32,6 +32,7 @@
 
 /* gaim headers */
 #include "gaim.h"
+#include "core.h"
 #include "notify.h"
 #include "version.h"
 #include "util.h"
@@ -178,7 +179,7 @@ static void notify_cb(void *opdata, OtrlNotifyLevel level,
 	const char *accountname, const char *protocol, const char *username,
 	const char *title, const char *primary, const char *secondary)
 {
-    GaimNotifyMsgType gaimlevel;
+    GaimNotifyMsgType gaimlevel = GAIM_NOTIFY_MSG_ERROR;
 
     switch (level) {
 	case OTRL_NOTIFY_ERROR:
@@ -208,19 +209,17 @@ static void update_context_list_cb(void *opdata)
     otrg_ui_update_keylist();
 }
 
-/* Forward declaration because we need to use &ui_ops in this function */
-static void confirm_fingerprint_cb(OtrlUserState us, void *opdata,
+static void confirm_fingerprint_cb(void *opdata, OtrlUserState us,
 	const char *accountname, const char *protocol, const char *username,
-	OTRKeyExchangeMsg kem,
-	void (*response_cb)(OtrlUserState us, OtrlMessageAppOps *ops,
-	    void *opdata, OTRConfirmResponse *response_data, int resp),
-	OTRConfirmResponse *response_data);
+	unsigned char fingerprint[20])
+{
+    otrg_dialog_unknown_fingerprint(us, accountname, protocol, username,
+	    fingerprint);
+}
 
 static void write_fingerprints_cb(void *opdata)
 {
-    gchar *storefile = g_build_filename(gaim_user_dir(), STOREFNAME, NULL);
-    otrl_privkey_write_fingerprints(otrg_plugin_userstate, storefile);
-    g_free(storefile);
+    otrg_plugin_write_fingerprints();
 }
 
 static void gone_secure_cb(void *opdata, ConnContext *context)
@@ -263,17 +262,6 @@ static OtrlMessageAppOps ui_ops = {
     log_message_cb
 };
 
-static void confirm_fingerprint_cb(OtrlUserState us, void *opdata,
-	const char *accountname, const char *protocol, const char *username,
-	OTRKeyExchangeMsg kem,
-	void (*response_cb)(OtrlUserState us, OtrlMessageAppOps *ops,
-	    void *opdata, OTRConfirmResponse *response_data, int resp),
-	OTRConfirmResponse *response_data)
-{
-    otrg_dialog_unknown_fingerprint(us, accountname, protocol, username, kem,
-	    response_cb, &ui_ops, opdata, response_data);
-}
-
 static void process_sending_im(GaimAccount *account, char *who, char **message,
 	void *m)
 {
@@ -313,11 +301,13 @@ static void process_sending_im(GaimAccount *account, char *who, char **message,
  * context, from the given account.  [account is actually a
  * GaimAccount*, but it's declared here as void* so this can be passed
  * as a callback.] */
-void otrg_plugin_send_default_query(ConnContext *context, void *account)
+void otrg_plugin_send_default_query(ConnContext *context, void *vaccount)
 {
-    char *msg = otrl_proto_default_query_msg(context->accountname);
-    otrg_plugin_inject_message((GaimAccount *)account, context->username,
-	    msg ? msg : "?OTR?");
+    GaimAccount *account = vaccount;
+    char *msg = otrl_proto_default_query_msg(context->accountname,
+	    otrg_ui_find_policy(account, context->username));
+    otrg_plugin_inject_message(account, context->username,
+	    msg ? msg : "?OTRv2?");
     free(msg);
 }
 
@@ -333,8 +323,9 @@ void otrg_plugin_send_default_query_conv(GaimConversation *conv)
     accountname = gaim_account_get_username(account);
     username = gaim_conversation_get_name(conv);
     
-    msg = otrl_proto_default_query_msg(accountname);
-    otrg_plugin_inject_message(account, username, msg ? msg : "?OTR?");
+    msg = otrl_proto_default_query_msg(accountname,
+	    otrg_ui_find_policy(account, username));
+    otrg_plugin_inject_message(account, username, msg ? msg : "?OTRv2?");
     free(msg);
 }
 
@@ -373,15 +364,8 @@ static gboolean process_receiving_im(GaimAccount *account, char **who,
     tlv = otrl_tlv_find(tlvs, OTRL_TLV_DISCONNECTED);
     if (tlv) {
 	/* Notify the user that the other side disconnected. */
-
-	char *msg = g_strdup_printf("OTR: %s has closed his private "
-		"connection to you; you should do the same.", username);
-
-	if (msg) {
-	    otrg_dialog_display_otr_message(accountname, protocol,
-		    username, msg);
-	    g_free(msg);
-	}
+	otrg_dialog_finished(accountname, protocol, username);
+	otrg_ui_update_keylist();
     }
     
     otrl_tlv_free(tlvs);
@@ -411,6 +395,16 @@ static void process_connection_change(GaimConnection *conn, void *data)
     otrg_dialog_resensitize_all();
 }
 
+static void process_button_type_change(const char *name, GaimPrefType type,
+	gpointer value, gpointer data)
+{
+    /* If the user changes the style of the buttons at the bottom of the
+     * conversation window, gaim annoyingly removes all the buttons from
+     * the bbox, and reinserts its own.  So we need to reinsert our
+     * buttons as well. */
+    otrg_dialog_resensitize_all();
+}
+
 static void otr_options_cb(GaimBlistNode *node, gpointer user_data)
 {
     /* We've already checked GAIM_BLIST_NODE_IS_BUDDY(node) */
@@ -423,11 +417,21 @@ static void otr_options_cb(GaimBlistNode *node, gpointer user_data)
 static void supply_extended_menu(GaimBlistNode *node, GList **menu)
 {
     GaimBlistNodeAction *act;
+    GaimBuddy *buddy;
+    GaimAccount *acct;
+    const char *proto;
 
     if (!GAIM_BLIST_NODE_IS_BUDDY(node)) return;
 
-     act = gaim_blist_node_action_new("OTR Settings", otr_options_cb, NULL);
-     *menu = g_list_append(*menu, act);
+    /* Extract the account, and then the protocol, for this buddy */
+    buddy = (GaimBuddy *)node;
+    acct = buddy->account;
+    if (acct == NULL) return;
+    proto = gaim_account_get_protocol_id(acct);
+    if (!otrg_plugin_proto_supports_otr(proto)) return;
+
+    act = gaim_blist_node_action_new("OTR Settings", otr_options_cb, NULL);
+    *menu = g_list_append(*menu, act);
 }
 
 /* Disconnect a context, sending a notice to the other side, if
@@ -438,6 +442,89 @@ void otrg_plugin_disconnect(ConnContext *context)
 	    context->accountname, context->protocol, context->username);
 }
 
+/* Write the fingerprints to disk. */
+void otrg_plugin_write_fingerprints(void)
+{
+    gchar *storefile = g_build_filename(gaim_user_dir(), STOREFNAME, NULL);
+    otrl_privkey_write_fingerprints(otrg_plugin_userstate, storefile);
+    g_free(storefile);
+}
+
+/* Find the ConnContext appropriate to a given GaimConversation. */
+ConnContext *otrg_plugin_conv_to_context(GaimConversation *conv)
+{
+    GaimAccount *account;
+    char *username;
+    const char *accountname, *proto;
+    ConnContext *context;
+
+    account = gaim_conversation_get_account(conv);
+    accountname = gaim_account_get_username(account);
+    proto = gaim_account_get_protocol_id(account);
+    username = g_strdup(
+	    gaim_normalize(account, gaim_conversation_get_name(conv)));
+
+    context = otrl_context_find(otrg_plugin_userstate, username, accountname,
+	    proto, 0, NULL, NULL, NULL);
+    g_free(username);
+
+    return context;
+}
+
+/* Find the GaimConversation appropriate to the given ConnContext.  If
+ * one doesn't yet exist, create it if force_create is true. */
+GaimConversation *otrg_plugin_context_to_conv(ConnContext *context,
+	int force_create)
+{
+    GaimAccount *account;
+    GaimConversation *conv;
+
+    account = gaim_accounts_find(context->accountname, context->protocol);
+    if (account == NULL) return NULL;
+
+    conv = gaim_find_conversation_with_account(context->username, account);
+    if (conv == NULL && force_create) {
+	conv = gaim_conversation_new(GAIM_CONV_IM, account, context->username);
+    }
+
+    return conv;
+}
+
+/* What level of trust do we have in the privacy of this ConnContext? */
+TrustLevel otrg_plugin_context_to_trust(ConnContext *context)
+{
+    TrustLevel level = TRUST_NOT_PRIVATE;
+
+    if (context && context->msgstate == OTRL_MSGSTATE_ENCRYPTED) {
+	if (context->active_fingerprint->trust &&
+		context->active_fingerprint->trust[0] != '\0') {
+	    level = TRUST_PRIVATE;
+	} else {
+	    level = TRUST_UNVERIFIED;
+	}
+    } else if (context && context->msgstate == OTRL_MSGSTATE_FINISHED) {
+	level = TRUST_FINISHED;
+    }
+
+    return level;
+}
+
+/* Send the OTRL_TLV_DISCONNECTED packets when we're about to quit. */
+static void process_quitting(void)
+{
+    ConnContext *context = otrg_plugin_userstate->context_root;
+    while(context) {
+	ConnContext *next = context->next;
+	if (context->msgstate == OTRL_MSGSTATE_ENCRYPTED &&
+		context->protocol_version > 1) {
+	    otrg_plugin_disconnect(context);
+	}
+	context = next;
+    }
+}
+
+static guint button_type_cbid;
+
 static gboolean otr_plugin_load(GaimPlugin *handle)
 {
     gchar *privkeyfile = g_build_filename(gaim_user_dir(), PRIVKEYFNAME, NULL);
@@ -445,6 +532,7 @@ static gboolean otr_plugin_load(GaimPlugin *handle)
     void *conv_handle = gaim_conversations_get_handle();
     void *conn_handle = gaim_connections_get_handle();
     void *blist_handle = gaim_blist_get_handle();
+    void *core_handle = gaim_get_core();
 
     if (!privkeyfile || !storefile) {
 	g_free(privkeyfile);
@@ -465,6 +553,8 @@ static gboolean otr_plugin_load(GaimPlugin *handle)
 
     otrg_ui_update_fingerprint();
 
+    gaim_signal_connect(core_handle, "quitting", otrg_plugin_handle,
+	    GAIM_CALLBACK(process_quitting), NULL);
     gaim_signal_connect(conv_handle, "sending-im-msg", otrg_plugin_handle,
             GAIM_CALLBACK(process_sending_im), NULL);
     gaim_signal_connect(conv_handle, "receiving-im-msg", otrg_plugin_handle,
@@ -477,6 +567,9 @@ static gboolean otr_plugin_load(GaimPlugin *handle)
 	    GAIM_CALLBACK(process_connection_change), NULL);
     gaim_signal_connect(blist_handle, "blist-node-extended-menu",
 	    otrg_plugin_handle, GAIM_CALLBACK(supply_extended_menu), NULL);
+    button_type_cbid = gaim_prefs_connect_callback(
+	    "/gaim/gtk/conversations/button_type",
+	    process_button_type_change, NULL);
 
     gaim_conversation_foreach(otrg_dialog_new_conv);
 
@@ -488,11 +581,14 @@ static gboolean otr_plugin_unload(GaimPlugin *handle)
     void *conv_handle = gaim_conversations_get_handle();
     void *conn_handle = gaim_connections_get_handle();
     void *blist_handle = gaim_blist_get_handle();
+    void *core_handle = gaim_get_core();
 
     /* Clean up all of our state. */
     otrl_userstate_free(otrg_plugin_userstate);
     otrg_plugin_userstate = NULL;
 
+    gaim_signal_disconnect(core_handle, "quitting", otrg_plugin_handle,
+	    GAIM_CALLBACK(process_quitting));
     gaim_signal_disconnect(conv_handle, "sending-im-msg", otrg_plugin_handle,
             GAIM_CALLBACK(process_sending_im));
     gaim_signal_disconnect(conv_handle, "receiving-im-msg", otrg_plugin_handle,
@@ -505,12 +601,24 @@ static gboolean otr_plugin_unload(GaimPlugin *handle)
 	    GAIM_CALLBACK(process_connection_change));
     gaim_signal_disconnect(blist_handle, "blist-node-extended-menu",
 	    otrg_plugin_handle, GAIM_CALLBACK(supply_extended_menu));
+    gaim_prefs_disconnect_callback(button_type_cbid);
 
     gaim_conversation_foreach(otrg_dialog_remove_conv);
 
     return 1;
 }
 
+/* Return 1 if the given protocol supports OTR, 0 otherwise. */
+int otrg_plugin_proto_supports_otr(const char *proto)
+{
+    /* IRC is the only protocol we know of that OTR doesn't work on (its
+     * maximum message size is too small to fit a Key Exchange Message). */
+    if (proto && !strcmp(proto, "prpl-irc")) {
+	return 0;
+    }
+    return 1;
+}
+
 #ifdef USING_GTK
 
 static GaimGtkPluginUiInfo ui_info =
